/*-
 *
 * Thu Nov 25 16:52:42 EET 1993
 */

#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "stat.h"
#include "ioctl.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "domain.h"
#include "mbuf.h"
#include "time.h"
#include "kernel.h"
#include "errno.h"

#include "net/if.h"
#include "net/route.h"
#include "net/af.h"

#include "ax25/ax.h"
#include "ax25/ax25.h"
#include "ax25/ax_pcb.h"
#include "ax25/ax_var.h"

extern int ax_monitored;

static axreset(), newstate(), try_send(), revert_offset(), clean_acked();

u_long	lapb_sendspace = 512;
u_long	lapb_recvspace = 512;

#define AXSTATE_OPEN(s) ((s) >= AXS_OPEN)

#define NEXT(n)         (((n) + 1) & 7)
#define PREV(n)         (((n) - 1) & 7)

#define BELOW(n, v)     ((((n) - (v)) & 7) + 1 > 4)
#define ABOVE(n, v)     ((((n) - (v) - 1) & 7) < 4)

#define SENDWINDOW(axp) (4 - (((axp)->axp_vs - (axp)->axp_una) & 7))
#define PTHRESHOLD      3

#define lapb_lookup(head, ifp, faddr, laddr) \
		(ax_pcblookup((head), (ifp), (faddr), (laddr), 1))

#define axtoso(ax) ((struct socket *) ((ax)->axp_socket))
#define sotoax(so) ((struct axpcb *) ((so)->so_pcb))

#define soisreadable(so) (sorwakeup(so))
#define sofull(so)       (sbspace(&(so)->so_rcv) <= 0)

int ax_rttdefault = PR_SLOWHZ * 2;

int ax_lapbT2ticks = 1;
int ax_lapbT3ticks = PR_SLOWHZ * 60 * 10;

int ax_lapbT1backoffs[] = {
	2, 4, 8, 16, 32, 64, 64, 64, 64, 64, 64,
};

int ax_lapbT1retries = 6;
int ax_lapbT3retries = 6;

int ax_ticks;

static
t1start(a)
	struct axpcb *a;
{
	a->axp_N2 = 0;
	a->axp_T1 = 1 + a->axp_srtt + ax_lapbT1backoffs[a->axp_N2];
	a->axp_T3 = ax_lapbT3ticks;
}
static
t2start(a)
	struct axpcb *a;
{
	a->axp_T2 = ax_lapbT2ticks;
	a->axp_T3 = ax_lapbT3ticks;
}
static
t1restart(a)
	struct axpcb *a;
{
	a->axp_T1 = 1 + a->axp_srtt + ax_lapbT1backoffs[++a->axp_N2];
	a->axp_T3 = ax_lapbT3ticks;
}
static
t3restart(a)
	struct axpcb *a;
{
	a->axp_T3 = 1 + a->axp_srtt + ax_lapbT1backoffs[++a->axp_N2];
}
static
t1stop(a)
	struct axpcb *a;
{
	if (a->axp_T2 == 0)
		a->axp_N2 = 0;
	a->axp_T1 = 0;
	a->axp_T3 = ax_lapbT3ticks;
}
static
t2stop(a)
	struct axpcb *a;
{
	if (a->axp_T1 == 0)
		a->axp_N2 = 0;
	a->axp_T2 = 0;
	a->axp_T3 = ax_lapbT3ticks;
}

axsupervisory(ifp, dst, src, cmdrsp, ctrl)
	struct ifnet *ifp;
	struct ax_addr *dst, *src;
	u_char ctrl;
{
	return ax25_send(ifp, dst, src, cmdrsp, ctrl, 0, NULL);
}

axcommand(axp, ctrl)
	struct axpcb *axp;
	u_char ctrl;
{
	if (ctrl & PF)
		lapb_rttstart(axp);

	axp->axp_localbusy = (ctrl & ~PF) == RNR;

	return ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		COMMAND, ctrl | (axp->axp_vr << 5), 0, NULL);
}

axresponse(axp, ctrl)
	struct axpcb *axp;
	u_char ctrl;
{
	axp->axp_localbusy = (ctrl & ~PF) == RNR;

	return ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		RESPONSE, ctrl | (axp->axp_vr << 5), 0, NULL);
}

axunnumbered(axp, ctrl)
	struct axpcb *axp;
	u_char ctrl;
{
	return ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		RESPONSE, ctrl, 0, NULL);
}

/*
 * Slow timeout routine, called every 500 ms.
 */

lapb_slowtimo()
{
	register struct axpcb *axp, *axpnext;
	int s = splnet();

	++ax_ticks;

	for (axp = axb.axp_next; axp != &axb; axp = axpnext) {
		axpnext = axp->axp_next;

		if (axp->axp_T2 && --axp->axp_T2 == 0)
			lapb_delack(axp);
		else
		if (axp->axp_T1 && --axp->axp_T1 == 0)
			lapb_rexmit(axp);
		else
		if (axp->axp_T3 && --axp->axp_T3 == 0)
			lapb_keepalive(axp);
		else
		if (axtoso(axp)->so_snd.sb_cc > 0)
			if (try_send(axp, 0))
				t1start(axp);
	}
	splx(s);
}

/*
 * T2 expired.
 */
lapb_delack(axp)
	register struct axpcb *axp;
{
	if (ax_monitored)
		ax_event(axp, "delack");

	axresponse(axp, (sofull(axtoso(axp)) ? RNR : RR));
}

/*
 * T3 expired.
 */
lapb_keepalive(axp)
	register struct axpcb *axp;
{
	if (ax_monitored)
		ax_event(axp, "keepalive");

	if (axp->axp_N2 >= ax_lapbT3retries) {
		newstate(axp, AXS_CLOSED, ETIMEDOUT);
	} else {
		axcommand(axp, (sofull(axtoso(axp)) ? RNR : RR) | PF);
		t3restart(axp);
	}
}

static
axreset(axp)
	struct axpcb *axp;
{
	t1stop(axp);
	t2stop(axp);

	axp->axp_vr = axp->axp_vs = axp->axp_una = 0;
	axp->axp_polltime = 0;

	axp->axp_srtt = ax_rttdefault;

	axp->axp_pollsent = 0;
	axp->axp_pollresent = 0;
	axp->axp_rejsent = 0;
	axp->axp_remotebusy = 0;
	axp->axp_localbusy = 0;
}

static
newstate(axp, state, reason)
	struct axpcb *axp;
	int state, reason;
{
	struct socket *so = axtoso(axp);

	axp->axp_state = state;
	if (reason)
		so->so_error = reason;

	if (ax_debug > 4)
		ax_pcbprint(axp);

	switch (state) {
	case AXS_CLOSED:
		lapb_close(so);
		break;
	case AXS_OPENING:
		soisconnecting(so);
		break;
	case AXS_CLOSING:
		soisdisconnecting(so);
		break;
	case AXS_LISTEN:
		break;
	case AXS_OPEN:
		soisconnected(so);
		break;
	default:
		panic("newstate: unknown state");
	}
}

lapb_connect(axp)
	struct axpcb *axp;
{
	int error;

	axstat.axs_connattempt++;

	error = ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		COMMAND, SABM | PF, 0, NULL);
	if (error == 0) {
		if (axp->axp_state != AXS_OPENING) {
			t1start(axp);
			axp->axp_polltime = ax_ticks;
		} else {
			t1restart(axp);
			axp->axp_pollresent = 1;
		}
		t2stop(axp);
		axp->axp_srtt = ax_rttdefault;
		axp->axp_pollsent = 1;
		axp->axp_rejsent = 0;
		axp->axp_remotebusy = 0;
		axp->axp_localbusy = 0;
		axp->axp_vr = axp->axp_vs = axp->axp_una = 0;

		newstate(axp, AXS_OPENING, 0);
	}
	return (error);
}

lapb_disconnect(axp)
	struct axpcb *axp;
{
	int error = 0;

	if (!AXSTATE_OPEN(axp->axp_state))
		return (0);

	error = ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		COMMAND, DISC | PF, 0, NULL);

	if (error == 0) {
		if (axp->axp_state != AXS_CLOSING) {
			t1start(axp);
		} else {
			t1restart(axp);
		}
		t2stop(axp);
		sbflush(&axtoso(axp)->so_rcv);
		newstate(axp, AXS_CLOSING, 0);
	}
	return (error);
}

/*
 * T1 expired.
 */
lapb_rexmit(axp)
	register struct axpcb *axp;
{
	if (ax_monitored)
		ax_event(axp, "rexmit");

	switch (axp->axp_state) {

	case AXS_OPENING:
		if (axp->axp_N2 >= ax_lapbT1retries) {
			newstate(axp, AXS_CLOSED, ETIMEDOUT);
			break;
		}
		lapb_connect(axp);
		break;

	case AXS_CLOSING:
		if (axp->axp_N2 >= ax_lapbT1retries) {
			newstate(axp, AXS_CLOSED, ETIMEDOUT);
			break;
		}
		lapb_disconnect(axp);
		break;

	case AXS_OPEN:
		if (axp->axp_pollsent && axp->axp_N2 >= ax_lapbT1retries) {
			newstate(axp, AXS_CLOSED, ETIMEDOUT);
			break;
		}
		if (axp->axp_una != axp->axp_vs) {
			revert_offset(axp, PREV(axp->axp_vs));

			if (try_send(axp, 1) == 0)
				axcommand(axp, (sofull(axtoso(axp)) ? RNR : RR) | PF);

			t1restart(axp);
			axp->axp_pollresent = 1;
		}
		break;
	}
}

lapb_rttstart(axp)
	struct axpcb *axp;
{
	if (axp->axp_pollsent)
		axp->axp_pollresent = 1;
	else
		axp->axp_polltime = ax_ticks;

	axp->axp_pollsent = 1;
}

lapb_rttupdate(axp)
	struct axpcb *axp;
{
	if (axp->axp_pollsent && !axp->axp_pollresent) {
		/*
		 * Dont let retried polls
		 * enter into calculation.
		 */
		int rtt = ax_ticks - axp->axp_polltime;
		int srtt = axp->axp_srtt;

		axp->axp_lastrtt = rtt;
		axp->axp_srtt = (3 * srtt + rtt) / 4;

		if (ax_monitored)
			ax_event(axp, "rttupdate");
	}
	axp->axp_pollsent = 0;
	axp->axp_pollresent = 0;
}

static
try_send(axp, sendpoll)
	struct axpcb *axp;
	int sendpoll;
{
	struct socket *so = axtoso(axp);
	struct mbuf *m;
	u_char ctrl;
	int error, len, win;

	if (axp->axp_remotebusy)
		return (0);

	len = so->so_snd.sb_cc - axp->axp_off;
	win = SENDWINDOW(axp);

	if (len <= 0 || win <= 0)
		return (0);
	if (len > PACLEN)
		len = PACLEN;
	if (axp->axp_pklengths[axp->axp_vs])
		len = axp->axp_pklengths[axp->axp_vs];

	MGETHDR(m, M_DONTWAIT, MT_HEADER);
	if (m == NULL)
		return (0);

	m->m_pkthdr.rcvif = axp->axp_ifp;
	m->m_data += max_linkhdr;
	m->m_len = 0;

	if (ax_debug > 2)
		printf("send offset %d, length %d, sb_cc %d\n",
			axp->axp_off, len, so->so_snd.sb_cc);

	if (len <= MHLEN - max_linkhdr) {
		m_copydata(so->so_snd.sb_mb, axp->axp_off, len, mtod(m, caddr_t));
		m->m_len += len;
	} else {
		m->m_next = m_copy(so->so_snd.sb_mb, axp->axp_off, len);
		if (m->m_next == 0) {
			m_freem(m);
			return (0);
		}
	}
	m->m_pkthdr.len = len;

	ctrl = INFO | (axp->axp_vr << 5) | (axp->axp_vs << 1);
	if (win <= 2 || sendpoll)
		ctrl |= PF;

	error = ax25_send(axp->axp_ifp, &axp->axp_faddr, &axp->axp_laddr,
		COMMAND, ctrl, AXTYPE_NOL3, m);
	if (error)
		return (0);

	if (ctrl & PF)
		lapb_rttstart(axp);

	axp->axp_pklengths[axp->axp_vs] = len;
	axp->axp_vs = NEXT(axp->axp_vs);
	axp->axp_off += len;

	return (1);
}

static
revert_offset(axp, nr)
	struct axpcb *axp;
	char nr;
{
	char vs = axp->axp_vs;

	if (ax_debug) {
		printf("revert_offset nr %d vs %d una %d off %d -> ",
			nr, vs, axp->axp_una, axp->axp_off);
	}
	while (BELOW(nr, vs)) {
		vs = PREV(vs);
#if 1
		if (axp->axp_off < axp->axp_pklengths[vs]) {
			printf("revert_offset: going negative!\n");
			axp->axp_off = 0;
		} else
#endif
		axp->axp_off -= axp->axp_pklengths[vs];
	}
	axp->axp_una = axp->axp_vs = vs;

	if (ax_debug) {
		printf(" vs %d una %d off %d\n",
			vs, axp->axp_una, axp->axp_off);
	}
}

static
clean_acked(axp, nr, sendreply)
	struct axpcb *axp;
	char nr;
	int sendreply;
{
	struct socket *so = axtoso(axp);
	char una = axp->axp_una;
	int acked = 0;

	while (ABOVE(nr, una)) {
#if 1
		if (axp->axp_off < axp->axp_pklengths[una]) {
			printf("clean_acked: going negative!\n");
			axp->axp_off = 0;
		} else
#endif
		axp->axp_off -= axp->axp_pklengths[una];
		acked += axp->axp_pklengths[una];
		axp->axp_pklengths[una] = 0;
		una = NEXT(una);
	}
	axp->axp_una = una;

	if (acked) {
		if (ax_debug) {
			printf("clean_acked nr %d, acked %d, una %d, off %d\n",
				nr, acked, axp->axp_una, axp->axp_off);
		}
		if (acked > so->so_snd.sb_cc) {
			printf("clean_acked acked %d, sb_cc %d\n",
				acked, so->so_snd.sb_cc);
			acked = so->so_snd.sb_cc;
		}
		sbdrop(&so->so_snd, acked);
		if (so->so_snd.sb_flags & SB_NOTIFY)
			sowwakeup(so);
	}
	if (sendreply) {
		axresponse(axp, (sofull(so) ? RNR : RR) | PF);
	}
}

lapb_input(ifp, dst, src, digis, ctrl, pid, m)
	struct ifnet *ifp;
	struct ax_addr *dst, *src, *digis;
	u_char ctrl, pid;
	struct mbuf *m;
{
	struct socket *so;
	struct axpcb *axp;
	u_char type, cmd, poll, final;
	char ns, nr;

	/*
	 * Version 1 protocol ?
	 */
	if (((dst->ax_ssid ^ src->ax_ssid) & SSID_C) == 0) {
		axstat.axs_oldproto++;
		m_freem(m);
		return;
	}
	if (cmd = (dst->ax_ssid & SSID_C)) {
		poll = (ctrl & PF);
		final =  0;
	} else {
		poll =  0;
		final = (ctrl & PF);
	}
	switch (ctrl & 3) {
	default:
		type = INFO;
		ns = (ctrl >> 1) & 7;
		nr = (ctrl >> 5) & 7;
		break;
	case (RR & 3):
		type = ctrl & 0xf;
		nr = (ctrl >> 5) & 7;
		break;
	case (SABM & 3):
		type = ctrl & ~PF;
		break;
	}
	if ((axp = lapb_lookup(&axb, ifp, src, dst)) == NULL) {
		if (poll)
			axsupervisory(ifp, src, dst, RESPONSE, DM | PF);
		m_freem(m);
		axstat.axs_notours++;
		return;
	}
	if (final) {
		/*
		 * Check rtt.
		 */
		lapb_rttupdate(axp);
	}
	so = axtoso(axp);

	switch (axp->axp_state) {

	case AXS_CLOSED:
		if (poll)
			axunnumbered(axp, DM | PF);
		break;

	case AXS_CLOSING:
		if (type == UA || type == DM) {
			/*
			 * This is finis.
			 */
			newstate(axp, AXS_CLOSED, 0);
			if (poll)
				axunnumbered(axp, DM | PF);
			break;
		}
		break;

	case AXS_LISTEN:
		/*
		 * Listening socket.
		 * Accept only SABMs.
		 */
		if (type == SABM) {
			/*
			 * Accept the new connection.
			 */
			so = sonewconn(so, 0);
			if (so == 0) {
				if (poll)
					axunnumbered(axp, DM | PF);
				break;
			}
			axp = sotoax(so);
			axp->axp_ifp = ifp;
			axp->axp_laddr = *dst;
			axp->axp_faddr = *src;

			newstate(axp, AXS_OPEN, 0);

			if (poll)
				axunnumbered(axp, UA | PF);
		} else {
			if (poll)
				axunnumbered(axp, DM | PF);
		}
		break;

	case AXS_OPENING:
		switch (type) {
		default:
			/*
			 * Funny frame in this state.
			 * Send DM.
			 */
			axunnumbered(axp, DM | (poll ? PF : 0));
			newstate(axp, AXS_CLOSED, ECONNRESET);
			break;

		case INFO:
		case RR:
			/*
			 * Maybe we lost his UA,
			 */
			lapb_connect(axp);
			break;

		case SABM:
			if (poll)
				axunnumbered(axp, UA | PF);
		case UA:
			newstate(axp, AXS_OPEN, 0);
			if (axp->axp_una == axp->axp_vs)
				t1stop(axp);
			if (try_send(axp, 0))
				t1start(axp);
			break;

		case FRMR:
		case DISC:
			if (poll)
				axunnumbered(axp, DM | PF);
			newstate(axp, AXS_CLOSED, ECONNABORTED);
			break;
		case DM:
			newstate(axp, AXS_CLOSED, ECONNREFUSED);
			break;
		}
		break;

	case AXS_OPEN:
		switch (type) {
		default:
			/*
			 * What is this ?
			 */
			break;

		case DM:
			if (poll)
				axunnumbered(axp, UA | PF);
			newstate(axp, AXS_CLOSED, ECONNRESET);
			break;

		case SABM:
			axp->axp_remotebusy = 0;
			if (poll)
				axunnumbered(axp, UA | PF);
			soisconnected(so);
			break;

		case FRMR:
			if (poll)
				axunnumbered(axp, UA | PF);
			newstate(axp, AXS_CLOSED, ECONNABORTED);
			break;

		case DISC:
			if (poll)
				axunnumbered(axp, UA | PF);
			newstate(axp, AXS_CLOSED, 0);
			break;

		case REJ:
			axp->axp_remotebusy = 0;
			clean_acked(axp, nr, poll);
			revert_offset(axp, nr);
			t1stop(axp);
			if (try_send(axp, 1))
				t1start(axp);
			break;

		case RR:
			axp->axp_remotebusy = 0;
			clean_acked(axp, nr, poll);
			if (final && axp->axp_una != axp->axp_vs)
				revert_offset(axp, nr);
			if (axp->axp_una == axp->axp_vs)
				t1stop(axp);
			if (try_send(axp, 0))
				t1start(axp);
			break;

		case RNR:
			axp->axp_remotebusy = 1;
			clean_acked(axp, nr, poll);
			if (final && axp->axp_una != axp->axp_vs)
				revert_offset(axp, nr);
			t1stop(axp);
			break;

		case INFO:
			clean_acked(axp, nr, 0);

			if (ns == axp->axp_vr) {
				axp->axp_rejsent = 0;
				if (!sofull(so)) {
					sbappend(&so->so_rcv, m);
					m = NULL;
					soisreadable(so);
					axp->axp_vr = NEXT(ns);
				}
				if (poll)
					axresponse(axp, (sofull(so) ? RNR : RR) | PF);
				else if (sofull(so))
					axcommand(axp, RNR);
			} else {
				u_char reply;

				if (axp->axp_rejsent) {
					reply = (sofull(so) ? RNR : RR);
				} else {
					axp->axp_rejsent = 1;
					reply = REJ;
				}
				if (poll)
					axresponse(axp, reply | PF);
				else
					axcommand(axp, reply);
			}
			if (try_send(axp, 0))
				t1start(axp);
			break;
		}
		break;

	}
	if (m)
		m_freem(m);
}

/*ARGSUSED*/
lapb_usrreq(so, req, m, nam, control)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *control;
{
	register struct axpcb *axp;
	int s, error = 0;

	if (req == PRU_CONTROL)
		return (ax_control(so, (int)m, (caddr_t)nam, (struct ifnet *)control));

	if (control && control->m_len) {
		m_freem(control);
		if (m)
			m_freem(m);
		return (EINVAL);
	}

	s = splnet();
	axp = sotoax(so);

	if (axp == NULL && req != PRU_ATTACH) {
		splx(s);
		return (EINVAL);		/* XXX */
	}
	switch (req) {

	case PRU_ATTACH:
		if (axp) {
			error = EISCONN;
			break;
		}
		error = lapb_attach(so);
		if (error)
			break;
		if ((so->so_options & SO_LINGER) && so->so_linger == 0)
			so->so_linger = AX25_LINGERTIME;
		break;

	case PRU_DETACH:
		error = lapb_detach(so);
		break;

	case PRU_BIND:
		error = ax_pcbbind(axp, nam);
		break;

	case PRU_LISTEN:
		if (axp->axp_ifp == NULL) {
			error = ax_pcbbind(axp, NULL);
			if (error)
				break;
		}
		axp->axp_state = AXS_LISTEN;
		break;

	case PRU_CONNECT:
		error = ax_pcbconnect(axp, nam);
		if (error)
			break;
		error = lapb_connect(axp);
		break;

	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	case PRU_DISCONNECT:
		if (axp->axp_ifp == NULL) {
			error = ENOTCONN;
			break;
		}
		try_send(axp, 0);
		error = lapb_disconnect(axp);
		break;

	case PRU_ACCEPT:
		ax_setpeeraddr(axp, nam);
		break;

	case PRU_SHUTDOWN:
		socantsendmore(so);
		break;

	case PRU_RCVD:
		if (axp->axp_localbusy)
			axcommand(axp, RR);
		break;

	case PRU_SEND:
		if (axp->axp_ifp == NULL) {
			error = ENOTCONN;
			break;
		}
		sbappend(&so->so_snd, m);
		break;

	case PRU_ABORT:
		lapb_drop(axp, ECONNABORTED);
		break;

	case PRU_SENSE:
		((struct stat *) m)->st_blksize = so->so_snd.sb_hiwat;
		break;

	case PRU_RCVOOB:
		error = EINVAL;
		break;

	case PRU_SENDOOB:
		error = EOPNOTSUPP;
		break;

	case PRU_SOCKADDR:
		ax_setsockaddr(axp, nam);
		break;

	case PRU_PEERADDR:
		ax_setpeeraddr(axp, nam);
		break;

	default:
		panic("lapb_usrreq: unknown request");
	}
	splx(s);
	return (error);
}

lapb_ctloutput(op, so, level, optname, mp)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **mp;
{
	int error = 0;
	struct axpcb *axp = sotoax(so);
	register struct mbuf *m;

#ifdef notyet
	if (level != ourproto)
		return (lower_ctloutput(op, so, level, optname, mp));
#endif

	switch (op) {

	case PRCO_SETOPT:
		m = *mp;
		switch (optname) {

		default:
			error = EINVAL;
			break;
		}
		if (m)
			(void) m_free(m);
		break;

	case PRCO_GETOPT:
		*mp = m = m_get(M_WAIT, MT_SOOPTS);
		m->m_len = sizeof(int);

		switch (optname) {
		default:
			error = EINVAL;
			break;
		}
		break;
	}
	return (error);
}

lapb_attach(so)
	struct socket *so;
{
	register struct axpcb *axp;
	int error;

	if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
		error = soreserve(so, lapb_sendspace, lapb_recvspace);
		if (error)
			return (error);
	}
	error = ax_pcballoc(so, &axb);
	if (error)
		return (error);

	axp = sotoax(so);
	axp->axp_state = AXS_CLOSED;
	axp->axp_srtt = ax_rttdefault;
	axp->axp_T1 = 0;
	axp->axp_T2 = 0;
	axp->axp_T3 = 0;

	return (0);
}

lapb_detach(so)
	struct socket *so;
{
	register struct axpcb *axp = sotoax(so);
	int error = 0;

	if (AXSTATE_OPEN(axp->axp_state))
		lapb_disconnect(axp);
	else
		lapb_close(so);

	return (error);
}

lapb_close(so)
	struct socket *so;
{
	soisdisconnected(so);
	ax_pcbdetach(sotoax(so));

	axstat.axs_closed++;
}

lapb_drop(axp)
	struct axpcb *axp;
{
	soisdisconnected(axtoso(axp));
	ax_pcbdetach(axp);

	axstat.axs_drops++;
}

