
#include "standard.h"

const word word_bitmask[] = {
	0x0001, 0x0002, 0x0004, 0x0008,
	0x0010, 0x0020, 0x0040, 0x0080,
	0x0100, 0x0200, 0x0400, 0x0800,
	0x1000, 0x2000, 0x4000, 0x8000,
};

const byte byte_bitmask[] = {
	0x01, 0x02, 0x04, 0x08,
	0x10, 0x20, 0x40, 0x80,
};

void
doh(char *what)
{
	green_leds(~0);

	print_string("\n\rdoh: ");
	print_string(what);

	SCI_drain();
	reboot(1);
}

#if 0 /* libc.s */

word
abs(short i)
{
	return ((i >= 0) ? i : -i);
}

STATIC long long0;

dword
labs(long i)
{
	return ((i >= 0) ? i : long0 - i);
}

long
lnegate(long i)
{
	return (long0 - i);
}

#endif /* libc.s */

#if 0 /* string.s */

int
strcmp(byte *a, byte *b)
{
	int i = 0;

	while (a[i] && a[i] == b[i] && ++i)
		;

	return (a[i] - b[i]);
}

int
strncmp(byte *a, byte *b, int size)
{
	int i = 0;

	while (size && a[i] && a[i] == b[i] && ++i)
		--size;

	return (size ? (a[i] - b[i]) : 0);
}

byte *
strcpyx(byte *dst, byte *src)
{
	while (*dst++ = *src++)
		;

	return (dst - 1);
}

bool
isdigit(byte c)
{
	return (c >= '0' && c <= '9');
}

byte
toupper(byte c)
{
	if (c >= 'a' && c <= 'z')
		c ^= 0x20;

	return (c);
}

byte
tolower(byte c)
{
	if (c >= 'A' && c <= 'Z')
		c ^= 0x20;

	return (c);
}

int
strcasecmp(byte *a, byte *b)
{
	int i = 0;

	while (a[i] && toupper(a[i]) == toupper(b[i]) && ++i)
		;

	return (toupper(a[i]) - toupper(b[i]));
}

void
strupper(byte *p)
{
	for ( ; *p; ++p)
		*p = toupper(*p);
}

void
strlower(byte *p)
{
	for ( ; *p; ++p)
		*p = tolower(*p);
}

int
strlen(byte *p0)
{
	byte *p = p0;

	while (*p)
		++p;

	return (p - p0);
}

void
strncpyx(byte *dst, byte *src, int size)
{
	if (size) {
		dst[--size] = 0;

		while (size--)
			if (*dst++ = *src)
				++src;
	}
}

void
bzero(void *vdst, int cc)
{
	byte *dst = vdst;

	if (cc) {
		do {
			*dst++ = 0;
		} while (--cc);
	}
}

void
bcopy(const void *vsrc, void *vdst, int cc)
{
	const byte *src = vsrc;
	byte *dst = vdst;

	if (cc) {
		if (dst < src) {
			do {
				*dst++ = *src++;
			} while (--cc);
		} else {
			dst += cc;
			src += cc;
			do {
				*--dst = *--src;
			} while (--cc);
		}
	}
}

int
bcmp(const void *vsrc, const void *vdst, int cc)
{
	const byte *src = vsrc;
	const byte *dst = vdst;

	if (cc) {
		do {
			if (*dst++ != *src++)
				break;
		} while (--cc);
	}
	return (cc);    /* 0 if no difference seen */
}

#endif /* of routines in string.s */

bool
isalnum(byte c)
{
	return ((c >= 'A' && c <= 'Z')
	     || (c >= 'a' && c <= 'z')
	     || isdigit(c));
}

void
strcat(byte *dst, byte *src)
{
	strcpyx(dst + strlen(dst), src);
}

void
print_byte(byte x)
{
	putchar("0123456789ABCDEF"[(x >> 4) & 0xF]);
	putchar("0123456789ABCDEF"[(x >> 0) & 0xF]);
}

#if 0 /* in string.s, silly nano-optimization */

void
print_string(const byte *s)
{
	byte c;

	while (c = *s++)
		putchar(c);
}

#endif

void
eol()
{
	print_string("\n\r");
}

void
blank()
{
	putchar(' ');
}

void
print_short(word x)
{
	if (x >= 10)
		print_short(x / 10);

	putchar((x % 10) + '0');
}

void
print_02u(word x)
{
	putchar(x / 10 + '0');
	putchar(x % 10 + '0');
}

void
print_signed_short(short x)
{
	if (x > 0) {
		putchar('+');
		print_short(x);
		return;
	}
	if (x < 0) {
		putchar('-');
		print_short(-x);
		return;
	}
	putchar(' ');
	putchar('0');
}

void
print_short_N(word x, short width)
{
	if (width > 1)
		print_short_N(x / 10, width - 1);

	putchar((x % 10) + '0');
}

void
print_signed_long(long *px)
{
	byte *p, buf[12];

	if (*px < 0) {
		dword x;

		putchar('-');
		x = lnegate(*px);
		sprint_010Lu(buf, &x);
	} else
		sprint_010Lu(buf, px);

	p = buf;
	while (*p == '0' && p[1])
		p++;
	while (*p)
		putchar(*p++);
}

void
print_word(word x)
{
	print_byte(x >> 8);
	print_byte(x >> 0);
}

void
print_dword(dword *dwp)
{
	byte *p = (void *) dwp;

	print_byte(p[0]);
	print_byte(p[1]);
	print_byte(p[2]);
	print_byte(p[3]);
}

void
print_buf(const byte *buf, int cc)
{
	while (cc--)
		print_byte(*buf++);
}

void
vis(const byte *buf, int cc)
{
	for ( ; cc--; ++buf)
		if (*buf <= ' ' || *buf >= 0x7F || *buf == '\\') {
			print_string("\\x");
			print_byte(*buf);
		} else
			putchar(*buf);
}

word
ascii_num(byte *s, int len)
{
	word n = 0;

	do {
		byte c = *s++;

		c -= '0';
		if (c >= 10)
			return (0xFFFF);

		n *= 10;
		n += c;
	} while (--len);

	return (n);
}

word
ascii_hex(byte *s, int len)
{
	word x = 0;

	do {
		byte c = *s++;

		if (c >= '0' && c <= '9')
			c -= '0';
		else
		if (c >= 'A' && c <= 'F')
			c -= 'A' - 0xA;
		else
		if (c >= 'a' && c <= 'f')
			c -= 'a' - 0xA;
		else
			return (0xFFFF);

		x <<= 4;
		x |= c;
	} while (--len);

	return (x);
}

#if 0 /* libc.s */

int
atoi(const byte *s)
{
	int n = 0;
	byte c;

	while (*s == ' ' || *s == '0')
		s++;

	while (c = *s++) {
		c -= '0';
		if (c > 9)
			break;
		n *= 10;
		n += c;
	}
	return (n);
}

#endif /* libc.s */

#if 0 /* crc.s */

word
crc_ax25(byte *p, int len)
{
	word crc = 0xFFFF;
	unsigned b;

	do {
		b = *p++ | 0x100;
		do {
			crc ^= b & 1;
			if (crc & 1) {
				crc >>= 1;
				crc ^= 0x8408;
			} else
				crc >>= 1;

			b >>= 1;
		} while (b != 1);
	} while (--len);

	return (~crc);
}

#endif

#if 0 /* unused */

/*
 *  Format integer.
 *  bufsiz includes the \0 at end.
 *  return pointer to first digit (result is at end of buf[], maybe)
 */
char *
ltoa(char *buf, int width_plus_one, dword num)
{
	if (width_plus_one < 10 + 1)
		doh("ltoa");

	sprint_010Lu(buf, &num);

	while (*buf == '0' && buf[1])
		++buf;

	return (buf);
}

/* same thing, just word arg */

char *
itoa(char *buf, int width_plus_one, word num)
{
	char *p = buf + width_plus_one;

	*--p = 0;
	do {
		*--p = '0' + (num % 10);
		num /= 10;
	} while (num && p > buf);

	return (p);
}

#endif

#if 0 /* nobody uses yet */

/*
 *  "just good enough" printf
 *
 *  fmt is a stack parameter too, thanks to ... prototype
 */
int
printf(char *fmt, ...)
{
	byte c;
	int width;
	char *cp;
	bool islong, zeropad;
	unsigned cc = 0;

	union {
		unsigned n;
		void    *p;
	} *ap = (void *) &fmt;

	char buf[20]; /* any long */


	while (c = *fmt++) {
		if (c != '%') {
			putchar(c);
			continue;
		}
		++ap; /* offset it by 2 below if non-word arg */

		islong = zeropad = FALSE;
		width = 0;

		if (*fmt == '0')
			zeropad = TRUE;

		while (*fmt >= '0' && *fmt <= '9') {
			width *= 10;
			width += *fmt - '0';
			++fmt;
		}
		if (*fmt == 'l') {
			++fmt;
			islong = TRUE;
		}
		if ((c = *fmt++) == 0)
			break;

		switch (c) {

		case 'c':
			putchar(ap->n);
			break;

		case 's':
			print_string(ap->p);
			break;

		case 'd':
		case 'u':
			if (islong) {
				cp = ltoa(buf, sizeof(buf), *(long *) ap);
				++ap;
			} else {
				cp = itoa(buf, sizeof(buf), ap->n);
			}
			print_string(cp);
			break;

		case 'x':
		case 'X':
			if (width == 2)
				print_byte(ap->n);
			else {
				if (islong) {
					print_word(ap->n);
					++ap;
				}
				print_word(ap->n);
			}
			break;
		}
	}
	return (cc);
}

#endif /* of unused printf */

bool
wildmat(char *pat, char *s)
{
	for ( ; *pat; ++s, ++pat) {

		/* multiple ifs smaller than a switch */

		if (*pat == '*') {
			while (*++pat == '*') /* squash multiple stars */
				;
			if (*pat == 0)
				return (TRUE);    /* star at end matches anything */

			while (*s) {
				if (wildmat(pat, s++))  /* try rest of pattern */
					return (TRUE);
			}
			return (FALSE);
		}
		if (*pat == '?') {
			if (*s == 0)        /* question mark is any single character */
				return (FALSE);
		} else {
			if (*s != *pat)     /* pattern is nonzero inside the loop */
				return (FALSE);
		}
	}
	return (*s == 0);
}

long
EDIVx_signed(long x, word d, word *rem)
{
	if (x < 0)
		return (lnegate(EDIVx(lnegate(x), d, rem)));

	return (EDIVx(x, d, rem));
}

long
SCALE32_signed(long *px, long *scale)
{
	if (*px < 0) {
		long xx = lnegate(*px);

		return (lnegate(SCALE32(&xx, scale)));
	}
	return (SCALE32(px, scale));
}

